#!/usr/bin/env python

"""
Module to read EdgeTech xstar segy files.  No writing at this time.

These files are produced by the SunOS 5.5 based Xstar workstation that
is used on Neal W. Driscolls chirp fish that was in use at SIO from
2002-2005 and at WHOI before that.  Ask Neal for the fish details.

     Copyright (C) 2005  Kurt Schwehr

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

$Id: segy_drv_xstar.py,v 1.6 2005/06/02 23:42:45 schwehr Exp $

According to the manual, this file is compatible with software
revisions 4.2.3, 4.3, 4.4 and 6.0.  Data collected with the special
5.0 revision does no adhere to this stucture.


Here is the manuals description of the trace data:

  4.3.1.2.2 Trace Data

  The trace data constsis of 6400-bytes of data and 1552-byte of
  zeros.  The trace begins life as the recording from the receivers.

  Bytes 0241-6640  Integer No. 120-3319

  Correlated return (3200 Short integers) The first 724 shor integers
  (from 120 to 843) represent the displayed processed envelope where
  each value is within 0-1024 range.

  Bytes 6641-8192 Integer No. 3320-4095

  776 zeros

Here is what is done by Jeff Dingler/Jeff Babcock in C_ConvertXstar
via Henkart's sioseis:

sioseis << eof1
procs diskin xstar header diskoa end

diskin
    format edgetech
    ipath $1 end
end

xstar
   type 2 dummies 2 end
end


header
     fno 0 lno 999999 ftr 0 ltr 999999
     l3 = l1 l6 = l3 end
end

diskoa
  opath ${BASE}.segy end
end

end
eof1
####################################
# Use SIOSEIS to output lat, long, shot, speed, course, shot (.xykcs)
# l21 & l22 are longitude and latitude (integers)
# i63 & i64 are course and speed (integers)
# l1 is trace sequence number in the line (effectively shot #)

set LATLONFILE = ${BASE}.xykcs
set LLSFILE    = ${BASE}.yxscs  #temporary file

sioseis << ! > $LLSFILE
procs diskin prout end

diskin
    ipath ${BASE}.segy end
end

prout 
    indices l21 l22 i64 i63 l1 format (5(1X,F10)) fno 1 lno 999999 end    
end

end
######################################################################

So looking at a database output, I can start to see the real,imaginary traces:

sqlite bpsegy.db 'select traceNo,X/3600.0,Y/3600.0 FROM  segyTrace' | head -10


15790|-120.101111111111|34.4061111111111
15790|-120.101111111111|34.4061111111111
15791|-120.101111111111|34.4061111111111
15791|-120.101111111111|34.4061111111111
15792|-120.101111111111|34.4061111111111
15792|-120.101111111111|34.4061111111111
15793|-120.101111111111|34.4061111111111
15793|-120.101111111111|34.4061111111111
15794|-120.101111111111|34.4061111111111
15794|-120.101111111111|34.4061111111111


There are two each of each TraceNo (which is basically shot number)

For more info, see Paul Henkart's Seiseis XSTAR notes:

http://sioseis.ucsd.edu/xstar.html

"""

import segy_drv_rev1
import copy


# FIX: Make sure all of the chirp specific samples are handled.
from segy_drv_rev1 import structIntCodes

# File header stuff
from segy_drv_rev1 import textFileHeaderEntries
from segy_drv_rev1 import dataFormats
from segy_drv_rev1 import dataFormatStruct
from segy_drv_rev1 import traceSortingCodes
from segy_drv_rev1 import sweepTypeCodes


binaryHeaderEntries = copy.deepcopy(segy_drv_rev1.binaryHeaderEntries)
#                                    Bytes   Integer No
del binaryHeaderEntries['JobId']   # 001-004 1600-1601
del binaryHeaderEntries['ReelNo']  # 009-012 1604-1605
del binaryHeaderEntries['TracesPerEnsemble']
del binaryHeaderEntries['AuxTracesPerEnsemble']
del binaryHeaderEntries['SampleInterval']
del binaryHeaderEntries['OrigSampleInterval']
del binaryHeaderEntries['OrigSamplesPerTrace']

# Bytes 027-400  / Integer No. 1613-1799 are Not used
# EdgeTech manual page 4-9, Figure 4.3.1.1.2
del binaryHeaderEntries['SweepLen']
del binaryHeaderEntries['SweepType']
del binaryHeaderEntries['TraceSweepChannel']
del binaryHeaderEntries['SweepTaperLenStart']
del binaryHeaderEntries['SweepTaperLenEnd']
del binaryHeaderEntries['TaperType']
del binaryHeaderEntries['CorrelatedTraces']
del binaryHeaderEntries['BinGainRecovered']
del binaryHeaderEntries['AmpRecovMeth']
del binaryHeaderEntries['MeasurementSystem']
del binaryHeaderEntries['ImpulsePolarity']
del binaryHeaderEntries['VibPolarity']

######################################################################
# Trace header stuff
######################################################################
from segy_drv_rev1 import traceHdrSize
from segy_drv_rev1 import traceIdCode
from segy_drv_rev1 import dataUse
from segy_drv_rev1 import coordUnits
from segy_drv_rev1 import gainType
from segy_drv_rev1 import taperType
from segy_drv_rev1 import timeBasis
from segy_drv_rev1 import overTravel
from segy_drv_rev1 import traceUnits
from segy_drv_rev1 import transdUnits
from segy_drv_rev1 import srcTypeOrient
from segy_drv_rev1 import srcMeasUnits
#from segy_drv_rev1 import traceHeader

def traceIntToBytes(integerNo):
    """ Translate from 4.3.1.2.1 240 Byte Trace Identification Header"""
    assert(0<=integerNo)
    assert(200>integerNo)
    return(integerNo*2 + 1)
def b(i):
    return(i*2 + 1)

# This vender Trace Id is whacked
# Uses Integer No from table 4.3.1.2.1
# the b() call converts to the byte position
traceHeader = {
    # ShotNo is an alias to TraceNo, since TraceNo is really the shot number
    # WARNING: traceNumber is the position in the file, while TraceNo is the actual shot count
    # for xstart (TraceNumber != ShotNo) && (ShotNo==TraceNo)
    # TraceNo is in the wrong place
    'Shotpoint': [b(0),b(1)+1,'Trace (record) number (integer - 32bit)'],
    'TraceNo': [b(0),b(1)+1,'Trace (record) number (integer - 32bit)'],
    'TraceId': [b(14),b(14)+1,'Trace identification code: 1=seismic data'],
    # Supposed to be DataUse:
    'MatchFiltered': [b(17),b(17)+1,'1=raw matched filtered data; 2 = processed image'],
    # Supposed to be 1st half of DistanceCenter
    'DistFishBehindAnt': [b(18),b(18)+1,'Distance of fish behind antenna (cm)'],
    # Supposed to be the other half of DistanceCenter
    'DistFishStarboardAnt': [b(19),b(19)+1,'Distance from antenna to fish in starboard direction(cm)'],

    # Make it available as either...
    'X': [b(36),b(37)+1,'NMEA Long is called X in SEGY Rev 1'],
    'NMEALong': [b(36),b(37)+1,'NMEA Long (if int 44=1, meters; if int 44=2, seconds of arc, W is negative)'],

    'Y': [b(38),b(39)+1,'NMEA Lat is Y in SEGY Rev 1'],
    'NMEALat': [b(38),b(39)+1,'NMEA Lat (if int 44=1, meters; if int 44=2, seconds of arc, S is negative)'],

    'GrpX': [b(40),b(41)+1,'NMEA Long mm or .0001 min of arc is GrpX in Segy Rev1'],
    'NMEALongSmall': [b(40),b(41)+1,'NMEA Long mm if int 44=1 or 0.0001 min of arc if 44=2'],

    'GrpY': [b(42),b(42)+1,'NMEALatSmall is GrpY in Segy Rev1'],
    'NMEALatSmall': [b(42),b(42)+1,'NMEA Lat mm or 0.0001 min of arc'],

    # Finally something that matches
    'CoordUnits': [b(44),b(44)+1,'1=meters, 2=seconds of arc'],

    # Supposed to be WxVel
    'PulsePowerManuf': [b(45),b(45)+1,'% Pulse power set by manufacturer in _text.txt'],

    # Supposed to be SubWxVel
    'PulsePowerManuf': [b(46),b(46)+1,'% Pulse power set by user'],

    # Supposed to be SrcCorr
    'TriggerWidthHeaveComp': [b(49),b(49)+1,'Trigger width from heave compensator(samples)'],

    # Supposed to be LatTB
    'LatTB': [b(53),b(53)+1,'ChanBDelay should be LatTB in Segy Rev1'],

    'ChanBDelay': [b(53),b(53)+1,'Channel-B delay (samples) (SB-0408 only)'],

    'Delay': [b(54),b(54)+1,'Delay between transmission and start of sampling (ms) (Rev4.2 only)'],

    # Should be MuteStart
    'VideoDelay': [b(55),b(55)+1,'Video delay (Post-processed data only)'],

    # Yo, what's up with the zero padding?  What a waste.
    'TraceSamples': [b(57),b(57)+1,'Number of samples (3976 = 3200+776 zeroes'],

    'SampleInterval': [b(58),b(58)+1,'A/D sample interval (micro sec)'],
    'InstGainConst': [b(60),b(60)+1,'Gain of programmable amp: 1,2,4,8,16,32,64,128'],
    'Correlated': [b(62),b(62)+1,'Correlated: 2 = yes'],
    'SweepFreqStart': [b(63),b(63)+1,'Sweep start freq (Hz)'],
    'SweepFreqEnd':   [b(64),b(64)+1,'Sweep end freq (Hz)'],
    'SweepLen': [b(65),b(65)+1,'Sweep length (ms)'],
    'SweepType': [b(66),b(66)+1,'Sweep type: 1=linear'],
    'TaperType': [b(69),b(69)+1,'Taper type: 3 = other'],

    'AliasFiltFreq': [b(70),b(70)+1,'Sample freq/2 (Hz)'],

    'Year': [b(78),b(78)+1,'Computer Year'],
    'Day': [b(79),b(79)+1,'Computer day of year'],
    'Hour': [b(80),b(80)+1,'Computer hour'],
    'Min': [b(81),b(81)+1,'Computer minute'],
    'Sec': [b(82),b(82)+1,'Computer seconds'],
    'TimeBasis': [b(83),b(83)+1,'Time basis: 3=other'],

    # Should be GeopGrpRlSw1
    'MaxValChanB': [b(85),b(85)+1,'Maximum value of digitized return measured by A/D channel-B: Channel-B signal meter'],

    # Should e first half of Xensemble
    'NumZerosEnd': [b(90),b(90)+1,'Number of zeros at end of unprocessed trace = 4096-number of samples acquired in a trace'],

    # Should be part of Xensemble
    'Trigger': [b(91),b(91)+1,'Trigger (0=internal, 1=external)'],

    # Should be Yensemble
    'MarkNo': [b(92),b(92)+1,'Mark Number (0=no mark)'],

    # Should be 2nd half of Yensemble
    'NMEAhour': [b(93),b(93)+1,'NMEA0183 hour'],

    # Should be PostInLineNo
    'NMEAmin':  [b(94),b(94)+1,'NMEA0183 minutes'],

    # Should be 2nd half of PostInLineNo
    'NMEAsec':  [b(95),b(95)+1,'NMEA0183 seconds'],

    # Should be PostCrossLineNo
    'course':  [b(96),b(96)+1,'NMEA0183 course'],
    # Should be 2nd half of PostCrossLineNo
    'speed':  [b(97),b(97)+1,'NMEA0183 speed'],
    # Should be 1st half of shotpoint - talk about confusing!
    'NMEAday':  [b(98),b(98)+1,'NMEA0183 day'],
    # Should be 2st half of shotpoint - talk about confusing!
    'NMEAyear':  [b(99),b(99)+1,'NMEA0183 year'],
    
    # Should be ShotpointScaler
    'TraceScale': [b(100),b(101)+1,'Trace scale factor.  WARNING: float'],

    # Should be TransdConstMant
    'MaxValChanA': [b(102),b(102)+1,'Maximum value of digitized return measured by A/D channel-A: Channel-A signal meter'],

    # Should be 2nd half of TransdConstMant
    'SystemConst': [b(103),b(103)+1,'System constant in _text.txt'],

    # Is the vehicle ID going to change part way through a file???!??!
    # Should be TransdConstExp
    'VehicleId': [b(104),b(104)+1,'Vehicle ID'],

    # Should be TransdUnits, DevTrId, and TimeScaler
    # FIX: deal with char fields
    'SoftwareVer': [b(105),b(107)+1,'Software Version (6 characters)'],

    # Should be SrcTypeOrient, SrcEDir, SrcEDirTenths, SrcMeasExp,
    # and SrcMeasUnits.  233-240 are unassigned
    # FIX: deal with char fields
    'Annotation': [b(108),b(119)+1,'Annotation (24 characters)']
}

# FIX: get rid of these two lines when text can be handled
del traceHeader['SoftwareVer']
del traceHeader['Annotation']

######################################################################
# Put together all the tables
######################################################################
from segy_drv_rev1 import fileHeaderTables
#from segy_drv_rev1 import traceHeaderTables
traceHeaderTables = copy.deepcopy(segy_drv_rev1.traceHeaderTables)
del traceHeaderTables['DataUse']
del traceHeaderTables['TransdUnits']
del traceHeaderTables['SrcTypeOrient']
del traceHeaderTables['TraceUnits']
del traceHeaderTables['OverTravel']
del traceHeaderTables['GainType']
del traceHeaderTables['SrcMeasUnits']

######################################################################
# Pulse types for Neal Driscoll's Chirp.  Are these the same for others?
######################################################################

# Length (ms), startFreq (kHz), endFreq (kHz)
pulseTypes={
    'L': [50, 1  ,  6  ],
    'M': [35, 1  ,  6  ],
    'N': [20, 3  , 15  ],
    'O': [20, 2.5, 12.5],
    'P': [10, 1  ,  5.5],
    'Q': [10, 3.5, 15.5],
    'R': [10, 1  , 15  ],
    'S': [ 3, 1  , 15  ],
    'T': [30, 1  , 15  ]
    }


######################################################################
# UNIT TESTS
######################################################################

import unittest

class TestXstar(unittest.TestCase):
    """
    Test as much as we can of the xstar driver
    """
    def testIntToBytes(self):
        self.failUnless(1==traceIntToBytes(0))
        self.failUnless(3==traceIntToBytes(1))
        self.failUnless(29==traceIntToBytes(14)) # Trace Id code

    def testTraceHeaders(self):
        """Make sure all trace headers do not overlap"""
        status=[]
        for i in range(240+1):
            status.append(False)

        #print len(status), ' status: ',status
        keyList = {}    
        for key in traceHeader.keys():
            #keyList.append(key)
            keyList[key]=False
            #print key

        # These are duplicates
        del keyList['X']
        del keyList['Y']
        del keyList['GrpX']
        del keyList['GrpY']
        del keyList['LatTB']

        for key in keyList.keys():
            th = traceHeader[key]
            #print th
            #print range(th[0],th[1]+1)
            for i in range(th[0],th[1]+1):
                if False!=status[i]:
                    print "ERROR: overlapping key:", key," at ",i
                    print th
                self.failUnless(False==status[i])
                status[i]=True

# FIX: add unit test stuff
if __name__ == '__main__':
    print 'FIX: need testing code!'
    unittest.main()
    #print binaryHeaderEntries
    #print traceHeader
